@@ -4,13 +4,14 @@ import os
 import time
 from datetime import datetime
 from pathlib import Path
-from typing import BinaryIO, Set, List, Optional, Union
+from typing import BinaryIO, Set, Dict, List, Optional, Union
 from urllib.parse import urlparse
+from botocore.exceptions import NoCredentialsError, PartialCredentialsError, ClientError
 
 import boto3
 import requests
 
-from .dataset import Dataset, DatasetFile, UploadInfo, DatasetList
+from .dataset import Dataset, DatasetFile, UploadInfo, DatasetList, DatasetCommit
 import random
 import string
 
@@ -73,9 +74,11 @@ def get_dataset(dataset_id: str, token: Optional[str] = None) -> Optional[Datase
         )
         # print("Get Files with subpath", files_with_subpath)
         # filter the files overlapped with files_with_subpath
-        files_key_set = set([file.key for file in files])
+        files_key_set = set([file.key for file in files_with_subpath])
         for file in files:
+            # print("file", file)
             if file.key not in files_key_set:
+                # print("not in")
                 files_with_subpath.append(file)
         # print("files_with_subpath", files_with_subpath)
         # files.extend(files_with_subpath)
@@ -86,6 +89,14 @@ def get_dataset(dataset_id: str, token: Optional[str] = None) -> Optional[Datase
             short_description=response.json().get("ShortDescription"),
             folder_name=response.json().get("FolderName"),
             files=files_with_subpath,
+            commits=[
+                DatasetCommit(
+                    commit.get("_id"),
+                    commit.get("CommitMessage"),
+                    parse_datetime(commit.get("CreateDate")),
+                )
+                for commit in response.json().get("DatasetVersions")
+            ],
             created_at=parse_datetime(response.json().get("CreateDate")),
             updated_at=parse_datetime(response.json().get("UpdateDate")),
         )
@@ -94,7 +105,12 @@ def get_dataset(dataset_id: str, token: Optional[str] = None) -> Optional[Datase
         return None
 
 
-def _update_dataset(id: str, files: List[DatasetFile], token: Optional[str] = None):
+def _update_dataset(
+    id: str,
+    files: List[DatasetFile],
+    commit_message: Optional[str] = None,
+    token: Optional[str] = None,
+):
     """
     Updates the dataset with the given ID by adding or removing files.
 
@@ -111,14 +127,17 @@ def _update_dataset(id: str, files: List[DatasetFile], token: Optional[str] = No
             )
     url = f"{HEYWHALE_SITE}/api/datasets/{id}/files"
     headers = {"x-kesci-token": token, "x-kesci-resource": id}
+    data = {
+        "Files": [file.key for file in files],
+        "FilesStructure": [
+            {"Token": file.key, "SubPath": file.sub_path} for file in files
+        ],
+    }
+    if commit_message is not None:
+        data["CommitMessage"] = commit_message
     response = requests.put(
         url,
-        json={
-            "Files": [file.key for file in files],
-            "FilesStructure": [
-                {"Token": file.key, "SubPath": file.sub_path} for file in files
-            ],
-        },
+        json=data,
         headers=headers,
         timeout=10,
     )
@@ -260,7 +279,7 @@ def upload_file(
                 return
         pp = str(Path(path_in_dataset).parent)
         new_dataset_files.append(DatasetFile("", object_key, 0, _norm_path(pp)))
-    _update_dataset(id, new_dataset_files, token)
+    _update_dataset(id, new_dataset_files, token=token)
 
 
 def _norm_path(pp: str) -> str:
@@ -272,7 +291,7 @@ def _norm_path(pp: str) -> str:
     Returns:
         str: 带尾部"/"的相对路径
     """
-    print(f"input pp {pp}")
+    # print(f"input pp {pp}")
     # 强加一个 /，按照示例格式来，怕没加没处理好。
     if not pp.endswith("/"):
         pp = pp + "/"
@@ -340,7 +359,7 @@ def upload_folder(
         # skip update if no files in directory
         return
     # print(new_dataset_files)
-    _update_dataset(id, list(new_dataset_files), token)
+    _update_dataset(id, list(new_dataset_files), token=token)
 
 
 def _init_cache(cache_dir: Optional[str | Path] = None) -> Path:
@@ -364,6 +383,8 @@ def _download_single_file(
     download_url: str,
     id: str,
     filename: str,
+    etag: str,
+    commit: str,
     cache_dir: Path,
     local_dir: Optional[str | Path] = None,
 ) -> Path:
@@ -385,30 +406,27 @@ def _download_single_file(
         .joinpath("datasets")
         .joinpath(id)
         .joinpath("snapshots")
-        .joinpath("1")
+        .joinpath(commit)
         .joinpath(filename)
     )
     if local_dir is not None:
         slk_path = Path(local_dir).joinpath(filename)
     if slk_path.exists():
         return slk_path
-    file_path = (
-        Path(cache_dir).expanduser().joinpath("blobs").joinpath(filename + ".lock")
-    )
-    file_path.parent.mkdir(parents=True, exist_ok=True)
-    if file_path.exists():
-        raise ValueError(
-            f"File '{filename}' is being downloaded, please wait. If the downloading process is interrupted, please delete the lock file '{file_path}' manually."
+    blob_path = Path(cache_dir).expanduser().joinpath("blobs").joinpath(etag)
+    if not blob_path.exists():
+        file_path = (
+            Path(cache_dir).expanduser().joinpath("blobs").joinpath(etag + ".download")
         )
-
-    m = hashlib.md5()
-    with open(file_path, "wb") as file:
-        for chunk in response.iter_content(chunk_size=chunk_size):
-            m.update(chunk)
-            file.write(chunk)
-            # pbar.update(len(chunk))
-    etag = m.hexdigest()
-    blob_path = file_path.rename(file_path.parent.joinpath(etag))
+        file_path.parent.mkdir(parents=True, exist_ok=True)
+        if file_path.exists():
+            raise ValueError(
+                f"File '{filename}' is being downloaded, please wait. If the downloading process is interrupted, please delete the lock file '{file_path}' manually."
+            )
+        with open(file_path, "wb") as file:
+            for chunk in response.iter_content(chunk_size=chunk_size):
+                file.write(chunk)
+        blob_path = file_path.rename(file_path.parent.joinpath(etag))
     slk_path.parent.mkdir(parents=True, exist_ok=True)
     slk_path.symlink_to(blob_path)
     return slk_path
@@ -416,6 +434,7 @@ def _download_single_file(
 
 def download_dir(
     id: str,
+    commit: Optional[str] = None,
     cache_dir: Optional[str | Path] = None,
     local_dir: Optional[str | Path] = None,
     token: Optional[str] = None,
@@ -435,33 +454,78 @@ def download_dir(
                 "No token provided and 'MW_TOKEN' environment variable is not set."
             )
     cache_dir = _init_cache(cache_dir)
+    dataset_detail = get_dataset(id)
+    if dataset_detail is None:
+        raise ValueError(f"Dataset '{id}' not found.")
+    path_to_key: Dict[str, str] = dict()
+    # print("files", dataset_detail.files)
+    for file in dataset_detail.files:
+        # print("pathind ", file.path_in_dataset())
+        path_to_key[file.path_in_dataset()] = file.key
+    upload_info = _get_update_token(token)
+    if upload_info is None:
+        raise ValueError("Failed to get upload token.")
+    session = boto3.Session(
+        aws_access_key_id=upload_info.ak,
+        aws_secret_access_key=upload_info.sk,
+        aws_session_token=upload_info.token,
+        region_name=upload_info.region,
+    )
+    s3 = session.client("s3", endpoint_url=upload_info.endpoint)
+    bucket_name = upload_info.bucket
     url = f"{HEYWHALE_SITE}/api/datasets/{id}/downloadUrl"
     headers = {
         "x-kesci-token": token,
     }
-    response = requests.get(url, headers=headers, timeout=10)
+    if commit is  None:
+        commit = dataset_detail.commits[-1]._id
+    params = {"Version": commit}
+    response = requests.get(url, headers=headers, params=params, timeout=10)
     if response.status_code == 200:
         files = response.json().get("files")
         for file in files:
             download_url = file.get("Url")
-            # Parse the URL to get components
-            parsed_url = urlparse(download_url)
-            # Extract the path component
-            # 本来返回值应该有个Token和parsed_url.path等价的但是有时候是None
-            parts = parsed_url.path.split("/")
-            # Remove the first 3 parts and join the remaining parts back into a string
-            remaining_parts = parts[
-                4:
-            ]  # This will slice the list to skip the first 3 elements
-            # Join the remaining parts with "/" to form the final string
-            filepath_in_dataset = "/".join(remaining_parts)
-            _download_single_file(
-                download_url,
-                id,
-                filepath_in_dataset,
-                cache_dir,
-                local_dir,
-            )
+            filepath_in_dataset = file.get("Name") + file.get("Ext")
+            sub_path = file.get("SubPath")
+            if sub_path is not None:
+                filepath_in_dataset = os.path.join(sub_path, filepath_in_dataset)
+            key = path_to_key[filepath_in_dataset]
+            # get etag
+            try:
+                # print("Get etag bucket key", bucket_name, key)
+                response = s3.head_object(Bucket=bucket_name, Key=key)
+                # Extract the ETag from the response
+                etag = response["ETag"]
+                # print(f"Etag {etag}")
+                # trim quote
+                etag = etag[1:-1]
+                _download_single_file(
+                    download_url,
+                    id,
+                    filepath_in_dataset,
+                    etag,
+                    commit,
+                    cache_dir,
+                    local_dir,
+                )
+            except NoCredentialsError:
+                print("Error: AWS credentials not available.")
+            except PartialCredentialsError:
+                print("Error: Incomplete AWS credentials.")
+            except ClientError as e:
+                # The client error includes more details about the error
+                if e.response["Error"]["Code"] == "404":
+                    print(
+                        f"Error: The object '{key}' does not exist in the bucket '{bucket_name}'."
+                    )
+                elif e.response["Error"]["Code"] == "403":
+                    print(
+                        f"Error: Access to the object '{key}' in the bucket '{bucket_name}' is forbidden."
+                    )
+                else:
+                    print(f"Error: {e.response['Error']['Message']}")
+            except Exception as e:
+                print(f"An unexpected error occurred: {str(e)}")
     else:
         print(response.text)
         response.raise_for_status()
@@ -472,7 +536,7 @@ def download_dir(
             .joinpath("datasets")
             .joinpath(id)
             .joinpath("snapshots")
-            .joinpath("1")
+            .joinpath(commit)
         )
         return dir
     else:
@@ -483,6 +547,7 @@ def download_dir(
 def download_file(
     id: str,
     filename: str,
+    commit: Optional[str] = None,
     cache_dir: Optional[str | Path] = None,
     local_dir: Optional[str | Path] = None,
     token: Optional[str] = None,
@@ -505,17 +570,42 @@ def download_file(
                 "No token provided and 'MW_TOKEN' environment variable is not set."
             )
     cache_dir = _init_cache(cache_dir)
-    # dataset_detail = get_dataset(id)
+    dataset_detail = get_dataset(id)
+    if dataset_detail is None:
+        raise ValueError(f"Dataset '{id}' not found.")
+    path_to_key: Dict[str, str] = dict()
+    for file in dataset_detail.files:
+        path_to_key[file.path_in_dataset()] = file.key
+
+    if filename not in path_to_key:
+        raise ValueError(f"File '{filename}' not found in dataset '{id}'.")
+    upload_info = _get_update_token(token)
+    if upload_info is None:
+        raise ValueError("Failed to get upload token.")
+    session = boto3.Session(
+        aws_access_key_id=upload_info.ak,
+        aws_secret_access_key=upload_info.sk,
+        aws_session_token=upload_info.token,
+        region_name=upload_info.region,
+    )
+    s3 = session.client("s3", endpoint_url=upload_info.endpoint)
+    bucket_name = upload_info.bucket
+
     url = f"{HEYWHALE_SITE}/api/datasets/{id}/downloadUrl"
     headers = {
         "x-kesci-token": token,
     }
-    response = requests.get(url, headers=headers, timeout=10)
+    if commit is  None:
+        commit = dataset_detail.commits[-1]._id
+    params = {"Version": commit}
+    response = requests.get(url, headers=headers, params=params, timeout=10)
     download_url: str = ""
     if response.status_code == 200:
         files = response.json().get("files")
+        # print(f"files {files}")
         for file in files:
-            if file.get("Name") + file.get("Ext") == filename:
+            sub_path = file.get("SubPath") if file.get("SubPath") is not None else ""
+            if os.path.join(sub_path, file.get("Name")) + file.get("Ext") == filename:
                 download_url = file.get("Url")
                 break
         else:
@@ -523,7 +613,36 @@ def download_file(
     else:
         print(response.text)
         response.raise_for_status()
-    slk_path = _download_single_file(download_url, id, filename, cache_dir, local_dir)
+    key = path_to_key[filename]
+    # get etag
+    try:
+        response = s3.head_object(Bucket=bucket_name, Key=key)
+        # Extract the ETag from the response
+        etag = response["ETag"]
+        # trim etag
+        etag = etag[1:-1]
+        # print(f"Etag {etag}")
+        slk_path = _download_single_file(
+            download_url, id, filename, etag, commit, cache_dir, local_dir
+        )
+    except NoCredentialsError:
+        print("Error: AWS credentials not available.")
+    except PartialCredentialsError:
+        print("Error: Incomplete AWS credentials.")
+    except ClientError as e:
+        # The client error includes more details about the error
+        if e.response["Error"]["Code"] == "404":
+            print(
+                f"Error: The object '{key}' does not exist in the bucket '{bucket_name}'."
+            )
+        elif e.response["Error"]["Code"] == "403":
+            print(
+                f"Error: Access to the object '{key}' in the bucket '{bucket_name}' is forbidden."
+            )
+        else:
+            print(f"Error: {e.response['Error']['Message']}")
+    except Exception as e:
+        print(f"An unexpected error occurred: {str(e)}")
     return slk_path
 
 
@@ -557,7 +676,7 @@ def delete_file(
         else:
             print(f"delete file {file.path_in_dataset()}")
             continue
-    _update_dataset(id, new_dataset_files, token)
+    _update_dataset(id, new_dataset_files, token=token)
     return
 
 
@@ -565,6 +684,30 @@ def delete_folder():
     pass
 
 
+def create_commit(
+    id: str,
+    commit_message: str,
+    token: Optional[str] = None,
+) -> DatasetCommit:
+    if token is None:
+        token = os.getenv("MW_TOKEN")
+        if not token:
+            raise ValueError(
+                "No token provided and 'MW_TOKEN' environment variable is not set."
+            )
+    upload_info = _get_update_token(token)
+    if upload_info is None:
+        raise ValueError("Failed to get upload token.")
+    dataset = get_dataset(id, token)
+    if dataset is None:
+        raise ValueError(f"Dataset '{id}' not found.")
+    _update_dataset(id, dataset.files, commit_message, token)
+    dataset = get_dataset(id, token)
+    if dataset is None:
+        raise ValueError(f"Dataset '{id}' not found.")
+    return dataset.commits[-1]
+
+
 def create_dataset(
     title: str,
     short_description: Optional[str] = None,
